@@ -3,6 +3,8 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdint.h>
+#include <inttypes.h>
 #include <string.h>
 #include "jdupes.h"
 #include "jody_win_unicode.h"
@@ -21,6 +23,8 @@ extern void deletefiles(file_t *files, int prompt, FILE *tty)
   unsigned int number, sum, max, x;
   size_t i;
 
+  if (!files) return;
+
   groups = get_max_dupes(files, &max, NULL);
 
   max++;
@@ -61,7 +65,7 @@ extern void deletefiles(file_t *files, int prompt, FILE *tty)
         /* prompt for files to preserve */
         printf("Set %u of %u: keep which files? (1 - %u, [a]ll, [n]one)",
           curgroup, groups, counter);
-        if (ISFLAG(flags, F_SHOWSIZE)) printf(" (%ju byte%c each)", (uintmax_t)files->size,
+        if (ISFLAG(flags, F_SHOWSIZE)) printf(" (%" PRIuMAX " byte%c each)", (uintmax_t)files->size,
           (files->size != 1) ? 's' : ' ');
         printf(": ");
         fflush(stdout);
