@@ -3,14 +3,16 @@
 
 #include "jdupes.h"
 
-#if !defined NO_HARDLINKS || !defined NO_SYMLINKS
+/* Compile out the code if no linking support is built in */
+#if !(defined NO_HARDLINKS && defined NO_SYMLINKS)
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 #include "act_linkfiles.h"
 #include "jody_win_unicode.h"
-#if defined _WIN32 || defined __CYGWIN__
+#ifdef ON_WINDOWS
  #include "win_stat.h"
 #endif
 
@@ -212,6 +214,7 @@ extern void linkfiles(file_t *files, const int hard)
 
         /* Create the desired hard link with the original file's name */
         errno = 0;
+        success = 0;
 #ifdef ON_WINDOWS
  #ifdef UNICODE
         if (!M2W(srcfile->d_name, wname2)) {
@@ -223,7 +226,6 @@ extern void linkfiles(file_t *files, const int hard)
         if (CreateHardLink(dupelist[x]->d_name, srcfile->d_name, NULL) == TRUE) success = 1;
  #endif
 #else
-        success = 0;
         if (hard) {
           if (link(srcfile->d_name, dupelist[x]->d_name) == 0) success = 1;
  #ifdef NO_SYMLINKS
@@ -241,11 +243,14 @@ extern void linkfiles(file_t *files, const int hard)
  #endif /* NO_SYMLINKS */
 #endif /* ON_WINDOWS */
         if (success) {
-          if (!ISFLAG(flags, F_HIDEPROGRESS)) printf("%s %s\n", (hard ? "---->" : "-@@->"), dupelist[x]->d_name);
+          if (!ISFLAG(flags, F_HIDEPROGRESS)) {
+            printf("%s ", hard ? "---->" : "-@@->");
+            fwprint(stdout, dupelist[x]->d_name, 1);
+          }
         } else {
           /* The link failed. Warn the user and put the link target back */
           if (!ISFLAG(flags, F_HIDEPROGRESS)) {
-            printf("-//-> "); fwprint(stderr, dupelist[x]->d_name, 1);
+            printf("-//-> "); fwprint(stdout, dupelist[x]->d_name, 1);
           }
           fprintf(stderr, "warning: unable to link '"); fwprint(stderr, dupelist[x]->d_name, 0);
           fprintf(stderr, "' -> '"); fwprint(stderr, srcfile->d_name, 0);
