@@ -2,6 +2,8 @@
  * This file is part of jdupes; see jdupes.c for license information */
 
 #include <stdio.h>
+#include <stdint.h>
+#include <inttypes.h>
 #include "jdupes.h"
 #include "jody_win_unicode.h"
 #include "act_printmatches.h"
@@ -11,11 +13,13 @@ extern void printmatches(file_t * restrict files)
   file_t * restrict tmpfile;
   int printed = 0;
 
+  LOUD(fprintf(stderr, "act_printmatches: %p\n", files));
+
   while (files != NULL) {
     if (ISFLAG(files->flags, F_HAS_DUPES)) {
       printed = 1;
       if (!ISFLAG(flags, F_OMITFIRST)) {
-        if (ISFLAG(flags, F_SHOWSIZE)) printf("%jd byte%c each:\n", (intmax_t)files->size,
+        if (ISFLAG(flags, F_SHOWSIZE)) printf("%" PRIdMAX " byte%c each:\n", (intmax_t)files->size,
          (files->size != 1) ? 's' : ' ');
         fwprint(stdout, files->d_name, 1);
       }
@@ -24,14 +28,14 @@ extern void printmatches(file_t * restrict files)
         fwprint(stdout, tmpfile->d_name, 1);
         tmpfile = tmpfile->duplicates;
       }
-      if (files->next != NULL) printf("\n");
+      if (files->next != NULL) fwprint(stdout, "", 1);
 
     }
 
     files = files->next;
   }
 
-  if (printed == 0) fprintf(stderr, "No duplicates found.\n");
+  if (printed == 0) fwprint(stderr, "No duplicates found.", 1);
 
   return;
 }
