@@ -31,6 +31,7 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdint.h>
+#include <inttypes.h>
 #ifndef OMIT_GETOPT_LONG
  #include <getopt.h>
 #endif
@@ -53,6 +54,7 @@
 #include "act_printmatches.h"
 #include "act_summarize.h"
 
+
 /* Detect Windows and modify as needed */
 #if defined _WIN32 || defined __CYGWIN__
  const char dir_sep = '\\';
@@ -77,6 +79,7 @@ wchar_t wname[PATH_MAX];
 wchar_t wname2[PATH_MAX];
 wchar_t wstr[PATH_MAX];
 int out_mode = _O_TEXT;
+int err_mode = _O_TEXT;
  #define M2W(a,b) MultiByteToWideChar(CP_UTF8, 0, a, -1, (LPWSTR)b, PATH_MAX)
  #define W2M(a,b) WideCharToMultiByte(CP_UTF8, 0, a, -1, (LPSTR)b, PATH_MAX, NULL, NULL)
 #endif /* UNICODE */
@@ -97,12 +100,6 @@ struct winstat ws;
 struct stat s;
 #endif
 
-static uintmax_t excludesize = 0;
-static enum {
-  SMALLERTHAN,
-  LARGERTHAN
-} excludetype = SMALLERTHAN;
-
 /* Larger chunk size makes large files process faster but uses more RAM */
 #ifndef CHUNK_SIZE
  #define CHUNK_SIZE 32768
@@ -134,6 +131,32 @@ static size_t auto_chunk_size;
 /* For interactive deletion input */
 #define INPUT_SIZE 512
 
+/* Size suffixes - this gets exported */
+const struct size_suffix size_suffix[] = {
+  /* Byte (someone may actually try to use this) */
+  { "b", 1 },
+  { "k", 1024 },
+  { "kib", 1024 },
+  { "m", 1048576 },
+  { "mib", 1048576 },
+  { "g", (uint64_t)1048576 * 1024 },
+  { "gib", (uint64_t)1048576 * 1024 },
+  { "t", (uint64_t)1048576 * 1048576 },
+  { "tib", (uint64_t)1048576 * 1048576 },
+  { "p", (uint64_t)1048576 * 1048576 * 1024},
+  { "pib", (uint64_t)1048576 * 1048576 * 1024},
+  { "e", (uint64_t)1048576 * 1048576 * 1048576},
+  { "eib", (uint64_t)1048576 * 1048576 * 1048576},
+  /* Decimal suffixes */
+  { "kb", 1000 },
+  { "mb", 1000000 },
+  { "gb", 1000000000 },
+  { "tb", 1000000000000 },
+  { "pb", 1000000000000000 },
+  { "eb", 1000000000000000000 },
+  { NULL, 0 },
+};
+
 /* Assemble extension string from compile-time options */
 static const char *extensions[] = {
   #ifdef ON_WINDOWS
@@ -196,9 +219,21 @@ struct travdone {
 };
 static struct travdone *travdone_head = NULL;
 
+/* Exclusion tree head and static tag list */
+struct exclude *exclude_head = NULL;
+const struct exclude_tags exclude_tags[] = {
+  { "dir",	X_DIR },
+  { "size+",	X_SIZE_GT },
+  { "size+=",	X_SIZE_GTEQ },
+  { "size-=",	X_SIZE_LTEQ },
+  { "size-",	X_SIZE_LT },
+  { "size=",	X_SIZE_EQ },
+  { NULL, 0 },
+};
+
 /* Required for progress indicator code */
 static uintmax_t filecount = 0;
-static uintmax_t progress = 0, dir_progress = 0, dupecount = 0;
+static uintmax_t progress = 0, item_progress = 0, dupecount = 0;
 /* Number of read loops before checking progress indicator */
 #define CHECK_MINIMUM 256
 
@@ -223,8 +258,8 @@ static unsigned int max_depth = 0;
 /* File tree head */
 static filetree_t *checktree = NULL;
 
-/* Directory parameter position counter */
-static unsigned int user_dir_count = 1;
+/* Directory/file parameter position counter */
+static unsigned int user_item_count = 1;
 
 /* registerfile() direction options */
 enum tree_direction { NONE, LEFT, RIGHT };
@@ -341,8 +376,8 @@ static void update_progress(const char * const restrict msg, const int file_perc
   gettimeofday(&time2, NULL);
 
   if (progress == 0 || time2.tv_sec > time1.tv_sec) {
-    fprintf(stderr, "\rProgress [%ju/%ju, %ju pairs matched] %ju%%", progress, filecount,
-      dupecount, (progress * 100) / filecount);
+    fprintf(stderr, "\rProgress [%" PRIuMAX "/%" PRIuMAX ", %" PRIuMAX " pairs matched] %" PRIuMAX "%%",
+      progress, filecount, dupecount, (progress * 100) / filecount);
     if (file_percent > -1 && msg != NULL) {
       fprintf(stderr, "  (%s: %d%%)         ", msg, file_percent);
       did_fpct = 1;
@@ -412,7 +447,7 @@ extern inline int getfilestats(file_t * const restrict file)
   file->mode = ws.mode;
  #ifndef NO_HARDLINKS
   file->nlink = ws.nlink;
- #endif /* NO_HARDLINKS */
+ #endif
 #else
   if (stat(file->d_name, &s) != 0) return -1;
   file->inode = s.st_ino;
@@ -420,6 +455,9 @@ extern inline int getfilestats(file_t * const restrict file)
   file->device = s.st_dev;
   file->mtime = s.st_mtime;
   file->mode = s.st_mode;
+ #ifndef NO_HARDLINKS
+  file->nlink = s.st_nlink;
+ #endif
  #ifndef NO_PERMS
   file->uid = s.st_uid;
   file->gid = s.st_gid;
@@ -433,6 +471,90 @@ extern inline int getfilestats(file_t * const restrict file)
 }
 
 
+static void add_exclude(const char *option)
+{
+  char *opt, *p;
+  struct exclude *excl = exclude_head;
+  const struct exclude_tags *tags = exclude_tags;
+  const struct size_suffix *ss = size_suffix;
+
+  if (option == NULL) nullptr("add_exclude()");
+
+  LOUD(fprintf(stderr, "add_exclude '%s'\n", option);)
+
+  opt = string_malloc(strlen(option) + 1);
+  if (opt == NULL) oom("add_exclude option");
+  strcpy(opt, option);
+  p = opt;
+
+  while (*p != ':' && *p != '\0') p++;
+
+  /* Split tag string into *opt (tag) and *p (value) */
+  if (*p == ':') {
+    *p = '\0';
+    p++;
+  }
+
+  while (tags->tag != NULL && strcmp(tags->tag, opt) != 0) tags++;
+  if (tags->tag == NULL) goto bad_tag;
+
+  /* Check for a tag that requires a value */
+  if (tags->flags & XX_EXCL_DATA && *p == '\0') goto spec_missing;
+
+  /* *p is now at the value, NOT the tag string! */
+
+  if (exclude_head != NULL) {
+    /* Add to end of exclusion stack if head is present */
+    while (excl->next != NULL) excl = excl->next;
+    excl->next = string_malloc(sizeof(struct exclude) + strlen(p));
+    if (excl->next == NULL) oom("add_exclude alloc");
+    excl = excl->next;
+  } else {
+    /* Allocate exclude_head if no exclusions exist yet */
+    exclude_head = string_malloc(sizeof(struct exclude) + strlen(p));
+    if (exclude_head == NULL) oom("add_exclude alloc");
+    excl = exclude_head;
+  }
+
+  /* Set tag value from predefined tag array */
+  excl->flags = tags->flags;
+
+  /* Initialize the new exclude element */
+  excl->next = NULL;
+  if (excl->flags & XX_EXCL_OFFSET) {
+    /* Exclude uses a number; handle it with possible suffixes */
+    *(excl->param) = '\0';
+    /* Get base size */
+    if (*p < '0' || *p > '9') goto bad_size_suffix;
+    excl->size = strtoll(p, &p, 10);
+    /* Handle suffix, if any */
+    if (*p != '\0') {
+      while (ss->suffix != NULL && strcasecmp(ss->suffix, p) != 0) ss++;
+      if (ss->suffix == NULL) goto bad_size_suffix;
+      excl->size *= ss->multiplier;
+    }
+  } else {
+    /* Exclude uses string data; just copy it */
+    excl->size = 0;
+    strcpy(excl->param, p);
+  }
+
+  LOUD(fprintf(stderr, "Added exclude: tag '%s', data '%s', size %lld, flags %d\n", opt, excl->param, (long long)excl->size, excl->flags);)
+  string_free(opt);
+  return;
+
+spec_missing:
+  fprintf(stderr, "Exclude spec missing or invalid: -X spec:data\n");
+  exit(EXIT_FAILURE);
+bad_tag:
+  fprintf(stderr, "Invalid exclusion tag was specified\n");
+  exit(EXIT_FAILURE);
+bad_size_suffix:
+  fprintf(stderr, "Invalid -X size suffix specified; use B or KMGTPE[i][B]\n");
+  exit(EXIT_FAILURE);
+}
+
+
 extern int getdirstats(const char * const restrict name,
         jdupes_ino_t * const restrict inode, dev_t * const restrict dev)
 {
@@ -443,10 +565,12 @@ extern int getdirstats(const char * const restrict name,
   if (win_stat(name, &ws) != 0) return -1;
   *inode = ws.inode;
   *dev = ws.device;
+  if (!S_ISDIR(ws.mode)) return 1;
 #else
   if (stat(name, &s) != 0) return -1;
   *inode = s.st_ino;
   *dev = s.st_dev;
+  if (!S_ISDIR(s.st_mode)) return 1;
 #endif /* ON_WINDOWS */
   return 0;
 }
@@ -470,7 +594,13 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
     return -1;
   }
 
- /* Exclude files by permissions if requested */
+  /* Exclude based on -1/--one-file-system */
+  if (ISFLAG(flags, F_ONEFS) && (file1->device != file2->device)) {
+    LOUD(fprintf(stderr, "check_conditions: files ignored: not on same filesystem\n"));
+    return -1;
+  }
+
+   /* Exclude files by permissions if requested */
   if (ISFLAG(flags, F_PERMISSIONS) &&
           (file1->mode != file2->mode
 #ifndef NO_PERMS
@@ -497,11 +627,13 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
 
   /* Exclude files that are not the same size */
   if (file1->size > file2->size) {
-    LOUD(fprintf(stderr, "check_conditions: no match: size of file1 > file2 (%jd > %jd)\n", (intmax_t)file1->size, (intmax_t)file2->size));
+    LOUD(fprintf(stderr, "check_conditions: no match: size of file1 > file2 (%" PRIdMAX " > %" PRIdMAX ")\n",
+      (intmax_t)file1->size, (intmax_t)file2->size));
     return -1;
   }
   if (file1->size < file2->size) {
-    LOUD(fprintf(stderr, "check_conditions: no match: size of file1 < file2 (%jd < %jd)\n", (intmax_t)file1->size, (intmax_t)file2->size));
+    LOUD(fprintf(stderr, "check_conditions: no match: size of file1 < file2 (%" PRIdMAX " < %"PRIdMAX ")\n",
+      (intmax_t)file1->size, (intmax_t)file2->size));
     return 1;
   }
 
@@ -511,12 +643,98 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
 }
 
 
+/* Check for exclusion conditions for a single file (1 = fail) */
+static int check_singlefile(file_t * const restrict newfile)
+{
+  static char tempname[PATHBUF_SIZE * 2];
+  static char *tp = tempname;
+  int excluded;
+
+  if (newfile == NULL) nullptr("check_singlefile()");
+
+  LOUD(fprintf(stderr, "check_singlefile: checking '%s'\n", newfile->d_name));
+
+  /* Exclude hidden files if requested */
+  if (ISFLAG(flags, F_EXCLUDEHIDDEN)) {
+    strcpy(tp, newfile->d_name);
+    tp = basename(tp);
+    if (tp[0] == '.' && strcmp(tp, ".") && strcmp(tp, "..")) {
+      LOUD(fprintf(stderr, "check_singlefile: excluding hidden file (-A on)\n"));
+      return 1;
+    }
+  }
+
+  /* Get file information and check for validity */
+  const int i = getfilestats(newfile);
+  if (i || newfile->size == -1) {
+    LOUD(fprintf(stderr, "check_singlefile: excluding due to bad stat()\n"));
+    return 1;
+  }
+
+  if (!S_ISDIR(newfile->mode)) {
+    /* Exclude zero-length files if requested */
+    if (newfile->size == 0 && !ISFLAG(flags, F_INCLUDEEMPTY)) {
+    LOUD(fprintf(stderr, "check_singlefile: excluding zero-length empty file (-z not set)\n"));
+    return 1;
+  }
+
+    /* Exclude files based on exclusion stack size specs */
+    excluded = 0;
+    for (struct exclude *excl = exclude_head; excl != NULL; excl = excl->next) {
+      uint32_t sflag = excl->flags & XX_EXCL_SIZE;
+      if (
+           ((sflag == X_SIZE_EQ) && (newfile->size != excl->size)) ||
+           ((sflag == X_SIZE_LTEQ) && (newfile->size <= excl->size)) ||
+           ((sflag == X_SIZE_GTEQ) && (newfile->size >= excl->size)) ||
+           ((sflag == X_SIZE_GT) && (newfile->size > excl->size)) ||
+           ((sflag == X_SIZE_LT) && (newfile->size < excl->size))
+      ) excluded = 1;
+    }
+    if (excluded) {
+      LOUD(fprintf(stderr, "check_singlefile: excluding based on xsize limit (-x set)\n"));
+      return 1;
+    }
+  }
+
+#ifdef ON_WINDOWS
+  /* Windows has a 1023 (+1) hard link limit. If we're hard linking,
+   * ignore all files that have hit this limit */
+ #ifndef NO_HARDLINKS
+  if (ISFLAG(flags, F_HARDLINKFILES) && newfile->nlink >= 1024) {
+  #ifdef DEBUG
+    hll_exclude++;
+  #endif
+    LOUD(fprintf(stderr, "check_singlefile: excluding due to Windows 1024 hard link limit\n"));
+    return 1;
+  }
+ #endif /* NO_HARDLINKS */
+#endif /* ON_WINDOWS */
+  return 0;
+}
+
+
+static file_t *init_newfile(const size_t len, file_t * restrict * const restrict filelistp)
+{
+  file_t * const restrict newfile = (file_t *)string_malloc(sizeof(file_t));
+  if (!newfile) oom("init_newfile() file structure");
+  memset(newfile, 0, sizeof(file_t));
+  newfile->d_name = (char *)string_malloc(len);
+  if (!newfile->d_name) oom("init_newfile() filename");
+
+  newfile->next = *filelistp;
+  newfile->user_order = user_item_count;
+  newfile->size = -1;
+  newfile->duplicates = NULL;
+  return newfile;
+}
+
+
 /* Create a new traversal check object and initialize its values */
 static struct travdone *travdone_alloc(const jdupes_ino_t inode, const dev_t device)
 {
   struct travdone *trav;
 
-  LOUD(fprintf(stderr, "travdone_alloc(%jd, %jd)\n", (intmax_t)inode, (intmax_t)device);)
+  LOUD(fprintf(stderr, "travdone_alloc(%" PRIdMAX ", %" PRIdMAX ")\n", (intmax_t)inode, (intmax_t)device);)
 
   trav = (struct travdone *)string_malloc(sizeof(struct travdone));
   if (trav == NULL) {
@@ -532,22 +750,44 @@ static struct travdone *travdone_alloc(const jdupes_ino_t inode, const dev_t dev
 }
 
 
+/* Add a single file to the file tree */
+static inline file_t *grokfile(const char * const restrict name, file_t * restrict * const restrict filelistp)
+{
+  file_t * restrict newfile;
+
+  if (!name || !filelistp) nullptr("grokfile()");
+  LOUD(fprintf(stderr, "grokfile: '%s' %p\n", name, filelistp));
+
+  /* Allocate the file_t and the d_name entries */
+  newfile = init_newfile(strlen(name) + 2, filelistp);
+
+  strcpy(newfile->d_name, name);
+
+  /* Single-file [l]stat() and exclusion condition check */
+  if (check_singlefile(newfile) != 0) {
+    LOUD(fprintf(stderr, "grokfile: check_singlefile rejected file\n"));
+    string_free(newfile->d_name);
+    string_free(newfile);
+    return NULL;
+  }
+  return newfile;
+}
+
+
 /* Load a directory's contents into the file tree, recursing as needed */
 static void grokdir(const char * const restrict dir,
                 file_t * restrict * const restrict filelistp,
                 int recurse)
 {
   file_t * restrict newfile;
-#ifndef NO_SYMLINKS
-  static struct stat linfo;
-#endif
   struct dirent *dirinfo;
   static int grokdir_level = 0;
   static char tempname[PATHBUF_SIZE * 2];
   size_t dirlen;
   struct travdone *traverse;
-  jdupes_ino_t inode;
-  dev_t device;
+  int i, single = 0;
+  jdupes_ino_t inode, n_inode;
+  dev_t device, n_device;
 #ifdef UNICODE
   WIN32_FIND_DATA ffd;
   HANDLE hFind = INVALID_HANDLE_VALUE;
@@ -557,10 +797,12 @@ static void grokdir(const char * const restrict dir,
 #endif
 
   if (dir == NULL || filelistp == NULL) nullptr("grokdir()");
-  LOUD(fprintf(stderr, "grokdir: scanning '%s' (order %d)\n", dir, user_dir_count));
+  LOUD(fprintf(stderr, "grokdir: scanning '%s' (order %d, recurse %d)\n", dir, user_item_count, recurse));
 
   /* Double traversal prevention tree */
-  if (getdirstats(dir, &inode, &device) != 0) goto error_travdone;
+  i = getdirstats(dir, &inode, &device);
+  if (i < 0) goto error_travdone;
+
   if (travdone_head == NULL) {
     travdone_head = travdone_alloc(inode, device);
     if (travdone_head == NULL) goto error_travdone;
@@ -570,12 +812,12 @@ static void grokdir(const char * const restrict dir,
       if (traverse == NULL) nullptr("grokdir() traverse");
       /* Don't re-traverse directories we've already seen */
       if (inode == traverse->inode && device == traverse->device) {
-        LOUD(fprintf(stderr, "already seen dir '%s', skipping\n", dir);)
+        LOUD(fprintf(stderr, "already seen item '%s', skipping\n", dir);)
         return;
       } else if (inode > traverse->inode || (inode == traverse->inode && device > traverse->device)) {
         /* Traverse right */
         if (traverse->right == NULL) {
-          LOUD(fprintf(stderr, "traverse dir right '%s'\n", dir);)
+          LOUD(fprintf(stderr, "traverse item right '%s'\n", dir);)
           traverse->right = travdone_alloc(inode, device);
           if (traverse->right == NULL) goto error_travdone;
           break;
@@ -585,7 +827,7 @@ static void grokdir(const char * const restrict dir,
       } else {
         /* Traverse left */
         if (traverse->left == NULL) {
-          LOUD(fprintf(stderr, "traverse dir left '%s'\n", dir);)
+          LOUD(fprintf(stderr, "traverse item left '%s'\n", dir);)
           traverse->left = travdone_alloc(inode, device);
           if (traverse->left == NULL) goto error_travdone;
           break;
@@ -596,9 +838,20 @@ static void grokdir(const char * const restrict dir,
     }
   }
 
-  dir_progress++;
+  item_progress++;
   grokdir_level++;
 
+  /* if dir is actually a file, just add it to the file tree */
+  if (i == 1) {
+    newfile = grokfile(dir, filelistp);
+    if (newfile == NULL) {
+      LOUD(fprintf(stderr, "grokfile rejected '%s'\n", dir));
+      return;
+    }
+    single = 1;
+    goto add_single_file;
+  }
+
 #ifdef UNICODE
   /* Windows requires \* at the end of directory names */
   strncpy(tempname, dir, PATHBUF_SIZE * 2);
@@ -611,7 +864,7 @@ static void grokdir(const char * const restrict dir,
 
   LOUD(fprintf(stderr, "FindFirstFile: %s\n", dir));
   hFind = FindFirstFile((LPCWSTR)wname, &ffd);
-  if (hFind == INVALID_HANDLE_VALUE) { fprintf(stderr, "\nfile handle bad\n"); goto error_cd; }
+  if (hFind == INVALID_HANDLE_VALUE) { LOUD(fprintf(stderr, "\nfile handle bad\n")); goto error_cd; }
   LOUD(fprintf(stderr, "Loop start\n"));
   do {
     char * restrict tp = tempname;
@@ -630,162 +883,102 @@ static void grokdir(const char * const restrict dir,
 #endif /* UNICODE */
 
     LOUD(fprintf(stderr, "grokdir: readdir: '%s'\n", dirinfo->d_name));
-    if (strcmp(dirinfo->d_name, ".") && strcmp(dirinfo->d_name, "..")) {
-      if (!ISFLAG(flags, F_HIDEPROGRESS)) {
-        gettimeofday(&time2, NULL);
-        if (progress == 0 || time2.tv_sec > time1.tv_sec) {
-          fprintf(stderr, "\rScanning: %ju files, %ju dirs (in %u specified)",
-              progress, dir_progress, user_dir_count);
-        }
-        time1.tv_sec = time2.tv_sec;
+    if (!strcmp(dirinfo->d_name, ".") || !strcmp(dirinfo->d_name, "..")) continue;
+    if (!ISFLAG(flags, F_HIDEPROGRESS)) {
+      gettimeofday(&time2, NULL);
+      if (progress == 0 || time2.tv_sec > time1.tv_sec) {
+        fprintf(stderr, "\rScanning: %" PRIuMAX " files, %" PRIuMAX " dirs (in %u specified)",
+            progress, item_progress, user_item_count);
       }
+      time1.tv_sec = time2.tv_sec;
+    }
 
-      /* Assemble the file's full path name, optimized to avoid strcat() */
-      dirlen = strlen(dir);
-      d_name_len = strlen(dirinfo->d_name);
-      memcpy(tp, dir, dirlen+1);
-      if (dirlen != 0 && tp[dirlen-1] != dir_sep) {
-        tp[dirlen] = dir_sep;
-        dirlen++;
-      }
-      if (dirlen + d_name_len + 1 >= (PATHBUF_SIZE * 2)) goto error_overflow;
-      tp += dirlen;
-      memcpy(tp, dirinfo->d_name, d_name_len);
-      tp += d_name_len;
-      *tp = '\0';
-      d_name_len++;
+    /* Assemble the file's full path name, optimized to avoid strcat() */
+    dirlen = strlen(dir);
+    d_name_len = strlen(dirinfo->d_name);
+    memcpy(tp, dir, dirlen+1);
+    if (dirlen != 0 && tp[dirlen-1] != dir_sep) {
+      tp[dirlen] = dir_sep;
+      dirlen++;
+    }
+    if (dirlen + d_name_len + 1 >= (PATHBUF_SIZE * 2)) goto error_overflow;
+    tp += dirlen;
+    memcpy(tp, dirinfo->d_name, d_name_len);
+    tp += d_name_len;
+    *tp = '\0';
+    d_name_len++;
 
-      /* Allocate the file_t and the d_name entries */
-      newfile = (file_t *)string_malloc(sizeof(file_t));
-      if (!newfile) oom("grokdir() file structure");
-      newfile->d_name = (char *)string_malloc(dirlen + d_name_len + 2);
-      if (!newfile->d_name) oom("grokdir() filename");
+    /* Allocate the file_t and the d_name entries */
+    newfile = init_newfile(dirlen + d_name_len + 2, filelistp);
 
-      newfile->next = *filelistp;
-      newfile->user_order = user_dir_count;
-      newfile->size = -1;
-      newfile->device = 0;
-      newfile->inode = 0;
-      newfile->mtime = 0;
-      newfile->mode = 0;
-#ifdef ON_WINDOWS
- #ifndef NO_HARDLINKS
-      newfile->nlink = 0;
- #endif
-#endif
-#ifndef NO_PERMS
-      newfile->uid = 0;
-      newfile->gid = 0;
-#endif
-      newfile->filehash = 0;
-      newfile->filehash_partial = 0;
-      newfile->duplicates = NULL;
-      newfile->flags = 0;
+    tp = tempname;
+    memcpy(newfile->d_name, tp, dirlen + d_name_len);
 
-      tp = tempname;
-      memcpy(newfile->d_name, tp, dirlen + d_name_len);
+    /* Single-file [l]stat() and exclusion condition check */
+    if (check_singlefile(newfile) != 0) {
+      LOUD(fprintf(stderr, "grokdir: check_singlefile rejected file\n"));
+      string_free(newfile->d_name);
+      string_free(newfile);
+      continue;
+    }
 
-      if (ISFLAG(flags, F_EXCLUDEHIDDEN)) {
-        /* WARNING: Re-used tp here to eliminate a strdup() */
-        strncpy(tp, newfile->d_name, dirlen + d_name_len);
-        tp = basename(tp);
-        if (tp[0] == '.' && strcmp(tp, ".") && strcmp(tp, "..")) {
-          LOUD(fprintf(stderr, "grokdir: excluding hidden file (-A on)\n"));
+    /* Optionally recurse directories, including symlinked ones if requested */
+    if (S_ISDIR(newfile->mode)) {
+      if (recurse) {
+        /* --one-file-system */
+        if (ISFLAG(flags, F_ONEFS)
+            && (getdirstats(newfile->d_name, &n_inode, &n_device) == 0)
+            && (device != n_device)) {
+          LOUD(fprintf(stderr, "grokdir: directory: not recursing (--one-file-system)\n"));
           string_free(newfile->d_name);
           string_free(newfile);
           continue;
         }
-      }
-
-      /* Get file information and check for validity */
-      const int i = getfilestats(newfile);
-      if (i || newfile->size == -1) {
-        LOUD(fprintf(stderr, "grokdir: excluding due to bad stat()\n"));
-        string_free(newfile->d_name);
-        string_free(newfile);
-        continue;
-      }
-
-      /* Exclude zero-length files if requested */
-      if (!S_ISDIR(newfile->mode) && newfile->size == 0 && !ISFLAG(flags, F_INCLUDEEMPTY)) {
-        LOUD(fprintf(stderr, "grokdir: excluding zero-length empty file (-z not set)\n"));
-        string_free(newfile->d_name);
-        string_free(newfile);
-        continue;
-      }
-
-      /* Exclude files below --xsize parameter */
-      if (!S_ISDIR(newfile->mode) && ISFLAG(flags, F_EXCLUDESIZE)) {
-        if (
-            ((excludetype == SMALLERTHAN) && (newfile->size < (off_t)excludesize)) ||
-            ((excludetype == LARGERTHAN) && (newfile->size > (off_t)excludesize))
-        ) {
-          LOUD(fprintf(stderr, "grokdir: excluding based on xsize limit (-x set)\n"));
-          string_free(newfile->d_name);
-          string_free(newfile);
-          continue;
-        }
-      }
-
 #ifndef NO_SYMLINKS
-      /* Get lstat() information */
-      if (lstat(newfile->d_name, &linfo) == -1) {
-        LOUD(fprintf(stderr, "grokdir: excluding due to bad lstat()\n"));
-        string_free(newfile->d_name);
-        string_free(newfile);
-        continue;
-      }
-#endif
-
-      /* Windows has a 1023 (+1) hard link limit. If we're hard linking,
-       * ignore all files that have hit this limit */
-#ifdef ON_WINDOWS
- #ifndef NO_HARDLINKS
-      if (ISFLAG(flags, F_HARDLINKFILES) && newfile->nlink >= 1024) {
-  #ifdef DEBUG
-        hll_exclude++;
-  #endif
-        LOUD(fprintf(stderr, "grokdir: excluding due to Windows 1024 hard link limit\n"));
-        string_free(newfile->d_name);
-        string_free(newfile);
-        continue;
-      }
- #endif
-#endif
-      /* Optionally recurse directories, including symlinked ones if requested */
-      if (S_ISDIR(newfile->mode)) {
-#ifndef NO_SYMLINKS
-        if (recurse && (/*ISFLAG(flags, F_FOLLOWLINKS) ||*/ !S_ISLNK(linfo.st_mode))) {
-          LOUD(fprintf(stderr, "grokdir: directory: recursing (-r/-R)\n"));
+        else if (ISFLAG(flags, F_FOLLOWLINKS) || !ISFLAG(newfile->flags, F_IS_SYMLINK)) {
+          LOUD(fprintf(stderr, "grokdir: directory(symlink): recursing (-r/-R)\n"));
           grokdir(newfile->d_name, filelistp, recurse);
         }
 #else
-        if (recurse) {
+        else {
           LOUD(fprintf(stderr, "grokdir: directory: recursing (-r/-R)\n"));
           grokdir(newfile->d_name, filelistp, recurse);
         }
 #endif
-        LOUD(fprintf(stderr, "grokdir: directory: not recursing\n"));
-        string_free(newfile->d_name);
-        string_free(newfile);
+      } else { LOUD(fprintf(stderr, "grokdir: directory: not recursing\n")); }
+      string_free(newfile->d_name);
+      string_free(newfile);
+      continue;
+    } else {
+add_single_file:
+      /* Add regular files to list, including symlink targets if requested */
+#ifndef NO_SYMLINKS
+      if (!ISFLAG(newfile->flags, F_IS_SYMLINK) || (ISFLAG(newfile->flags, F_IS_SYMLINK) && ISFLAG(flags, F_FOLLOWLINKS))) {
+#else
+      if (S_ISREG(newfile->mode)) {
+#endif
+        *filelistp = newfile;
+        filecount++;
+        progress++;
+
       } else {
-        /* Add regular files to list, including symlink targets if requested */
-#ifndef NO_SYMLINKS
-        if (S_ISREG(linfo.st_mode) || (S_ISLNK(linfo.st_mode) && ISFLAG(flags, F_FOLLOWLINKS))) {
-#else
-        if (S_ISREG(newfile->mode)) {
-#endif
-          *filelistp = newfile;
-          filecount++;
-          progress++;
-        } else {
-          LOUD(fprintf(stderr, "grokdir: not a regular file: %s\n", newfile->d_name);)
-          string_free(newfile->d_name);
-          string_free(newfile);
+        LOUD(fprintf(stderr, "grokdir: not a regular file: %s\n", newfile->d_name);)
+        string_free(newfile->d_name);
+        string_free(newfile);
+        if (single == 1) {
+          single = 0;
+          goto skip_single;
         }
+        continue;
       }
     }
+    /* Skip directory stuff if adding only a single file */
+    if (single == 1) {
+      single = 0;
+      goto skip_single;
+    }
   }
+
 #ifdef UNICODE
   while (FindNextFile(hFind, &ffd) != 0);
   FindClose(hFind);
@@ -793,11 +986,11 @@ static void grokdir(const char * const restrict dir,
   closedir(cd);
 #endif
 
-
+skip_single:
   grokdir_level--;
   if (grokdir_level == 0 && !ISFLAG(flags, F_HIDEPROGRESS)) {
-    fprintf(stderr, "\rScanning: %ju files, %ju dirs (in %u specified)",
-            progress, dir_progress, user_dir_count);
+    fprintf(stderr, "\rScanning: %" PRIuMAX " files, %" PRIuMAX " items (in %u specified)",
+            progress, item_progress, user_item_count);
   }
   return;
 
@@ -812,6 +1005,7 @@ error_overflow:
   exit(EXIT_FAILURE);
 }
 
+
 /* Use Jody Bruchon's hash function on part or all of a file */
 static hash_t *get_filehash(const file_t * const restrict checkfile,
                 const size_t max_read)
@@ -824,7 +1018,7 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
   int check = 0;
 
   if (checkfile == NULL || checkfile->d_name == NULL) nullptr("get_filehash()");
-  LOUD(fprintf(stderr, "get_filehash('%s', %jd)\n", checkfile->d_name, (intmax_t)max_read);)
+  LOUD(fprintf(stderr, "get_filehash('%s', %" PRIdMAX ")\n", checkfile->d_name, (intmax_t)max_read);)
 
   /* Get the file size. If we can't read it, bail out early */
   if (checkfile->size == -1) {
@@ -850,10 +1044,11 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
     *hash = checkfile->filehash_partial;
     /* Don't bother going further if max_read is already fulfilled */
     if (max_read != 0 && max_read <= PARTIAL_HASH_SIZE) {
-      LOUD(fprintf(stderr, "Partial hash size (%d) >= max_read (%ju), not hashing anymore\n", PARTIAL_HASH_SIZE, (uintmax_t)max_read);)
+      LOUD(fprintf(stderr, "Partial hash size (%d) >= max_read (%" PRIuMAX "), not hashing anymore\n", PARTIAL_HASH_SIZE, (uintmax_t)max_read);)
       return hash;
     }
   }
+  errno = 0;
 #ifdef UNICODE
   if (!M2W(checkfile->d_name, wstr)) file = NULL;
   else file = _wfopen(wstr, FILE_MODE_RO);
@@ -861,7 +1056,7 @@ static hash_t *get_filehash(const file_t * const restrict checkfile,
   file = fopen(checkfile->d_name, FILE_MODE_RO);
 #endif
   if (file == NULL) {
-    fprintf(stderr, "\nerror opening file "); fwprint(stderr, checkfile->d_name, 1);
+    fprintf(stderr, "\n%s error opening file ", strerror(errno)); fwprint(stderr, checkfile->d_name, 1);
     return NULL;
   }
   /* Actually seek past the first chunk if applicable
@@ -1372,6 +1567,10 @@ static inline void help_text(void)
 {
   printf("Usage: jdupes [options] DIRECTORY...\n\n");
 
+#ifdef LOUD
+  printf(" -@ --loud        \toutput annoying low-level debug info while running\n");
+#endif
+  printf(" -1 --one-file-system \tdo not match files on different filesystems/devices\n");
   printf(" -A --nohidden    \texclude hidden files from consideration\n");
 #ifdef ENABLE_BTRFS
   printf(" -B --dedupe      \tSend matches to btrfs for block-level deduplication\n");
@@ -1382,6 +1581,9 @@ static inline void help_text(void)
   printf("                  \twith -s or --symlinks, or when specifying a\n");
   printf("                  \tparticular directory more than once; refer to the\n");
   printf("                  \tdocumentation for additional information\n");
+#ifdef DEBUG
+  printf(" -D --debug       \toutput debug statistics after completion\n");
+#endif
   printf(" -f --omitfirst   \tomit the first file in each set of matches\n");
   printf(" -h --help        \tdisplay this help message\n");
 #ifndef NO_HARDLINKS
@@ -1411,8 +1613,9 @@ static inline void help_text(void)
   printf(" -p --permissions \tdon't consider files with different owner/group or\n");
   printf("                  \tpermission bits as duplicates\n");
 #endif
-  printf(" -r --recurse     \tfor every directory given follow subdirectories\n");
-  printf("                  \tencountered within\n");
+  printf(" -Q --quick       \tskip byte-for-byte confirmation for quick matching\n");
+  printf("                  \tWARNING: -Q can result in data loss! Be very careful!\n");
+  printf(" -r --recurse     \tfor every directory, process its subdirectories too\n");
   printf(" -R --recurse:    \tfor each directory given after this option follow\n");
   printf("                  \tsubdirectories encountered within (note the ':' at\n");
   printf("                  \tthe end of the option, manpage for more details)\n");
@@ -1428,9 +1631,12 @@ static inline void help_text(void)
   printf(" -v --version     \tdisplay jdupes version and license information\n");
   printf(" -x --xsize=SIZE  \texclude files of size < SIZE bytes from consideration\n");
   printf("    --xsize=+SIZE \t'+' specified before SIZE, exclude size > SIZE\n");
-  printf("                  \tK/M/G size suffixes can be used (case-insensitive)\n");
+  printf(" -X --exclude=spec:info\texclude files based on specified criteria\n");
+  printf("                  \tspecs: dir size+-=\n");
+  printf("                  \tExclusions are cumulative: -X dir:abc -X dir:efg\n");
   printf(" -z --zeromatch   \tconsider zero-length files to be duplicates\n");
   printf(" -Z --softabort   \tIf the user aborts (i.e. CTRL-C) act on matches so far\n");
+  printf("\nFor sizes, K/M/G/T/P/E[B|iB] suffixes can be used (case-insensitive)\n");
 #ifdef OMIT_GETOPT_LONG
   printf("Note: Long options are not supported in this build.\n\n");
 #endif
@@ -1447,7 +1653,7 @@ int main(int argc, char **argv)
   static file_t *files = NULL;
   static file_t *curfile;
   static char **oldargv;
-  static char *endptr;
+  static char *xs;
   static int firstrecurse;
   static int opt;
   static int pm = 1;
@@ -1457,43 +1663,39 @@ int main(int argc, char **argv)
   static const struct option long_options[] =
   {
     { "loud", 0, 0, '@' },
+    { "one-file-system", 0, 0, '1' },
     { "nohidden", 0, 0, 'A' },
     { "dedupe", 0, 0, 'B' },
     { "delete", 0, 0, 'd' },
     { "debug", 0, 0, 'D' },
     { "omitfirst", 0, 0, 'f' },
     { "help", 0, 0, 'h' },
-#ifndef NO_HARDLINKS
     { "hardlinks", 0, 0, 'H' },
-    { "linkhard", 0, 0, 'L' },
-#endif
     { "reverse", 0, 0, 'i' },
     { "isolate", 0, 0, 'I' },
+    { "linksoft", 0, 0, 'l' },
+    { "linkhard", 0, 0, 'L' },
     { "summarize", 0, 0, 'm'},
     { "summary", 0, 0, 'm' },
     { "noempty", 0, 0, 'n' },
     { "noprompt", 0, 0, 'N' },
     { "order", 1, 0, 'o' },
     { "paramorder", 0, 0, 'O' },
-#ifndef NO_PERMS
     { "permissions", 0, 0, 'p' },
-#endif
     { "quiet", 0, 0, 'q' },
     { "quick", 0, 0, 'Q' },
     { "recurse", 0, 0, 'r' },
     { "recursive", 0, 0, 'r' },
     { "recurse:", 0, 0, 'R' },
     { "recursive:", 0, 0, 'R' },
-#ifndef NO_SYMLINKS
-    { "linksoft", 0, 0, 'l' },
     { "symlinks", 0, 0, 's' },
-#endif
     { "size", 0, 0, 'S' },
     { "version", 0, 0, 'v' },
     { "xsize", 1, 0, 'x' },
+    { "exclude", 1, 0, 'X' },
     { "zeromatch", 0, 0, 'z' },
     { "softabort", 0, 0, 'Z' },
-    { 0, 0, 0, 0 }
+    { NULL, 0, 0, 0 }
   };
 #define GETOPT getopt_long
 #else
@@ -1512,9 +1714,13 @@ int main(int argc, char **argv)
   argv = (char **)string_malloc(sizeof(char *) * argc);
   if (!argv) oom("main() unicode argv");
   widearg_to_argv(argc, wargv, argv);
+  /* fix up __argv so getopt etc. don't crash */
+  __argv = argv;
   /* Only use UTF-16 for terminal output, else use UTF-8 */
-  if (!_isatty(_fileno(stdout))) out_mode = _O_U8TEXT;
+  if (!_isatty(_fileno(stdout))) out_mode = _O_BINARY;
   else out_mode = _O_U16TEXT;
+  if (!_isatty(_fileno(stderr))) err_mode = _O_BINARY;
+  else err_mode = _O_U16TEXT;
 #endif /* UNICODE */
 
   /* Auto-tune chunk size to be half of L1 data cache if possible */
@@ -1532,12 +1738,15 @@ int main(int argc, char **argv)
   oldargv = cloneargs(argc, argv);
 
   while ((opt = GETOPT(argc, argv,
-  "@ABdDfhHiIlLmnNOpqQrRsSvzZo:x:"
+  "@1ABdDfhHiIlLmnNOpqQrRsSvzZo:x:X:"
 #ifndef OMIT_GETOPT_LONG
           , long_options, NULL
 #endif
          )) != EOF) {
     switch (opt) {
+    case '1':
+      SETFLAG(flags, F_ONEFS);
+      break;
     case 'A':
       SETFLAG(flags, F_EXCLUDEHIDDEN);
       break;
@@ -1615,35 +1824,22 @@ int main(int argc, char **argv)
       SETFLAG(flags, F_SOFTABORT);
       break;
     case 'x':
-      SETFLAG(flags, F_EXCLUDESIZE);
+      fprintf(stderr, "-x/--xsize is deprecated; use -X size[+-=]:size[suffix] instead\n");
+      xs = string_malloc(8 + strlen(optarg));
+      if (xs == NULL) oom("xsize temp string");
+      strcat(xs, "size");
       if (*optarg == '+') {
-        excludetype = LARGERTHAN;
+        strcat(xs, "+:");
         optarg++;
+      } else {
+        strcat(xs, "-=:");
       }
-      excludesize = (uintmax_t)strtoull(optarg, &endptr, 0);
-      switch (*endptr) {
-        case 'k':
-        case 'K':
-          excludesize = excludesize * 1024;
-          endptr++;
-          break;
-        case 'm':
-        case 'M':
-          excludesize = excludesize * 1024 * 1024;
-          endptr++;
-          break;
-        case 'g':
-        case 'G':
-          excludesize = excludesize * 1024 * 1024 * 1024;
-          endptr++;
-          break;
-        default:
-          break;
-      }
-      if (*endptr != '\0') {
-        fprintf(stderr, "invalid value for --xsize: '%s'\n", optarg);
-        exit(EXIT_FAILURE);
-      }
+      strcat(xs, optarg);
+      add_exclude(xs);
+      string_free(xs);
+      break;
+    case 'X':
+      add_exclude(optarg);
       break;
     case '@':
 #ifdef LOUD_DEBUG
@@ -1660,7 +1856,8 @@ int main(int argc, char **argv)
       } else if (sizeof(uintptr_t) == 4) {
         if (sizeof(long) == 4) printf("32-bit\n");
         else if (sizeof(long) == 8) printf("32-bit i64\n");
-      } else printf("%ld-bit i%ld\n", sizeof(uintptr_t) * 8, sizeof(long) * 8);
+      } else printf("%u-bit i%u\n", (unsigned int)(sizeof(uintptr_t) * 8),
+          (unsigned int)(sizeof(long) * 8));
 
       printf("Compile-time extensions:");
       if (*extensions != NULL) {
@@ -1700,18 +1897,19 @@ int main(int argc, char **argv)
       break;
     case 'B':
 #ifdef ENABLE_BTRFS
-    SETFLAG(flags, F_DEDUPEFILES);
-    /* btrfs will do the byte-for-byte check itself */
-    SETFLAG(flags, F_QUICKCOMPARE);
-    /* It is completely useless to dedupe zero-length extents */
-    CLEARFLAG(flags, F_INCLUDEEMPTY);
+      SETFLAG(flags, F_DEDUPEFILES);
+      /* btrfs will do the byte-for-byte check itself */
+      SETFLAG(flags, F_QUICKCOMPARE);
+      /* It is completely useless to dedupe zero-length extents */
+      CLEARFLAG(flags, F_INCLUDEEMPTY);
 #else
-    fprintf(stderr, "This program was built without btrfs support\n");
-    exit(EXIT_FAILURE);
+      fprintf(stderr, "This program was built without btrfs support\n");
+      exit(EXIT_FAILURE);
 #endif
-    break;
+      break;
 
     default:
+      if (opt != '?') fprintf(stderr, "Sorry, using '-%c' is not supported in this build.\n", opt);
       fprintf(stderr, "Try `jdupes --help' for more information.\n");
       string_malloc_destroy();
       exit(EXIT_FAILURE);
@@ -1724,12 +1922,6 @@ int main(int argc, char **argv)
     exit(EXIT_FAILURE);
   }
 
-  if (ISFLAG(flags, F_ISOLATE) && optind == (argc - 1)) {
-    fprintf(stderr, "Isolation requires at least two directories on the command line\n");
-    string_malloc_destroy();
-    exit(EXIT_FAILURE);
-  }
-
   if (ISFLAG(flags, F_RECURSE) && ISFLAG(flags, F_RECURSEAFTER)) {
     fprintf(stderr, "options --recurse and --recurse: are not compatible\n");
     string_malloc_destroy();
@@ -1777,7 +1969,7 @@ int main(int argc, char **argv)
     for (int x = optind; x < firstrecurse; x++) {
       slash_convert(argv[x]);
       grokdir(argv[x], &files, 0);
-      user_dir_count++;
+      user_item_count++;
     }
 
     /* Set F_RECURSE for directories after --recurse: */
@@ -1786,19 +1978,22 @@ int main(int argc, char **argv)
     for (int x = firstrecurse; x < argc; x++) {
       slash_convert(argv[x]);
       grokdir(argv[x], &files, 1);
-      user_dir_count++;
+      user_item_count++;
     }
   } else {
     for (int x = optind; x < argc; x++) {
       slash_convert(argv[x]);
       grokdir(argv[x], &files, ISFLAG(flags, F_RECURSE));
-      user_dir_count++;
+      user_item_count++;
     }
   }
 
   if (ISFLAG(flags, F_REVERSESORT)) sort_direction = -1;
   if (!ISFLAG(flags, F_HIDEPROGRESS)) fprintf(stderr, "\n");
-  if (!files) exit(EXIT_SUCCESS);
+  if (!files) {
+    fwprint(stderr, "No duplicates found.", 1);
+    exit(EXIT_SUCCESS);
+  }
 
   curfile = files;
   progress = 0;
@@ -1922,13 +2117,14 @@ skip_file_scan:
     fprintf(stderr, "\n%d partial (+%d small) -> %d full hash -> %d full (%d partial elim) (%d hash%u fail)\n",
         partial_hash, small_file, full_hash, partial_to_full,
         partial_elim, hash_fail, (unsigned int)sizeof(hash_t)*8);
-    fprintf(stderr, "%ju total files, %ju comparisons, branch L %u, R %u, both %u\n",
+    fprintf(stderr, "%" PRIuMAX " total files, %" PRIuMAX " comparisons, branch L %u, R %u, both %u, max tree depth %u\n",
         filecount, comparisons, left_branch, right_branch,
-        left_branch + right_branch);
-    fprintf(stderr, "Max tree depth: %u; SMA: allocs %ju, free %ju, fail %ju, reuse %ju, scan %ju, tails %ju\n",
-        max_depth, sma_allocs, sma_free_good, sma_free_ignored,
-        sma_free_reclaimed, sma_free_scanned, sma_free_tails);
-    fprintf(stderr, "I/O chunk size: %ju KiB (%s)\n", (uintmax_t)(auto_chunk_size >> 10),
+        left_branch + right_branch, max_depth);
+    fprintf(stderr, "SMA: allocs %" PRIuMAX ", free %" PRIuMAX " (merge %" PRIuMAX ", repl %" PRIuMAX "), fail %" PRIuMAX ", reuse %" PRIuMAX ", scan %" PRIuMAX ", tails %" PRIuMAX "\n",
+        sma_allocs, sma_free_good, sma_free_merged, sma_free_replaced,
+	sma_free_ignored, sma_free_reclaimed,
+        sma_free_scanned, sma_free_tails);
+    fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (%s)\n", (uintmax_t)(auto_chunk_size >> 10),
         (pci.l1 + pci.l1d) != 0 ? "dynamically sized" : "default size");
 #ifdef ON_WINDOWS
  #ifndef NO_HARDLINKS
