@@ -8,21 +8,6 @@
 extern "C" {
 #endif
 
-#include <limits.h>
-#include <stdint.h>
-#include <sys/types.h>
-#include <unistd.h>
-#include "string_malloc.h"
-#include "jody_hash.h"
-#include "jody_sort.h"
-#include "version.h"
-
-/* Optional btrfs support */
-#ifdef ENABLE_BTRFS
-#include <sys/ioctl.h>
-#include <linux/btrfs.h>
-#endif
-
 /* Detect Windows and modify as needed */
 #if defined _WIN32 || defined __CYGWIN__
  #define ON_WINDOWS 1
@@ -37,6 +22,25 @@ extern "C" {
  #include "win_stat.h"
  #define S_ISREG WS_ISREG
  #define S_ISDIR WS_ISDIR
+#endif /* Win32 */
+
+#include <limits.h>
+#include <stdint.h>
+#include <sys/types.h>
+#include <unistd.h>
+#include "string_malloc.h"
+#include "jody_hash.h"
+#include "jody_sort.h"
+#include "version.h"
+
+/* Optional btrfs support */
+#ifdef ENABLE_BTRFS
+#include <sys/ioctl.h>
+#include <linux/btrfs.h>
+#endif
+
+/* Some types are different on Windows */
+#ifdef ON_WINDOWS
  typedef uint64_t jdupes_ino_t;
  typedef uint32_t jdupes_mode_t;
  extern const char dir_sep;
@@ -64,6 +68,7 @@ extern wchar_t wname[PATH_MAX];
 extern wchar_t wname2[PATH_MAX];
 extern wchar_t wstr[PATH_MAX];
 extern int out_mode;
+extern int err_mode;
  #define M2W(a,b) MultiByteToWideChar(CP_UTF8, 0, a, -1, (LPWSTR)b, PATH_MAX)
  #define W2M(a,b) WideCharToMultiByte(CP_UTF8, 0, a, -1, (LPSTR)b, PATH_MAX, NULL, NULL)
 #endif /* UNICODE */
@@ -133,6 +138,7 @@ extern uint_fast32_t flags;
 #define F_ISOLATE		0x00100000U
 #define F_MAKESYMLINKS		0x00200000U
 #define F_PRINTMATCHES		0x00400000U
+#define F_ONEFS			0x00800000U
 
 #define F_LOUD			0x40000000U
 #define F_DEBUG			0x80000000U
@@ -184,9 +190,11 @@ typedef struct _file {
   uid_t uid;
   gid_t gid;
 #endif
-#ifdef ON_WINDOWS
- #ifndef NO_HARDLINKS
-  DWORD nlink;
+#ifndef NO_HARDLINKS
+ #ifndef ON_WINDOWS
+  nlink_t nlink;
+ #else
+  uint32_t nlink;  /* link count on Windows is always a DWORD */
  #endif
 #endif
 } file_t;
@@ -209,6 +217,50 @@ extern struct winstat ws;
 extern struct stat s;
 #endif
 
+
+/* -X exclusion parameter stack */
+struct exclude {
+  struct exclude *next;
+  unsigned int flags;
+  int64_t size;
+  char param[];
+};
+
+/* Exclude parameter flags */
+#define X_DIR			0x00000001U
+#define X_SIZE_EQ		0x00000002U
+#define X_SIZE_GT		0x00000004U
+#define X_SIZE_LT		0x00000008U
+/* The X-than-or-equal are combination flags */
+#define X_SIZE_GTEQ		0x00000006U
+#define X_SIZE_LTEQ		0x0000000aU
+
+/* Size specifier flags */
+#define XX_EXCL_SIZE		0x0000000eU
+/* Flags that use numeric offset instead of a string */
+#define XX_EXCL_OFFSET		0x0000000eU
+/* Flags that require a data parameter */
+#define XX_EXCL_DATA		0x0000000fU
+
+/* Exclude definition array */
+struct exclude_tags {
+  const char * const tag;
+  const uint32_t flags;
+};
+
+extern const struct exclude_tags exclude_tags[];
+extern struct exclude *exclude_head;
+
+
+/* Suffix definitions (treat as case-insensitive) */
+struct size_suffix {
+  const char * const suffix;
+  const int64_t multiplier;
+};
+
+extern const struct size_suffix size_suffix[];
+
+
 extern void oom(const char * const restrict msg);
 extern void nullptr(const char * restrict func);
 extern int file_has_changed(file_t * const restrict file);
